from blacksheep.messages import Request
from blacksheep.contents import Content


def get_status_line(status: int) -> bytes: ...

def is_small_request(request: Request) -> bool: ...

def request_has_body(request: Request) -> bool: ...

def write_small_request(request: Request) -> bytes: ...

def write_request_without_body(request: Request) -> bytes: ...

# TODO: how to annotate an async generator?
async def write_chunks(content: Content): ...
